"""Race channel cleanup with creating a snoop channel.

This process attempts to hang up a channel (from the SIP end) at the
same time as requesting that a snoop channel is created for it.  There
are two acceptable outcomes: the resource service can return an error
indicating that it could not create a snoop channel for a channel that
has been hung up, or the successful creation and tear down of a snoop
channel.  In asterisk from 18.0 to current, one possible behaviour is
for asterisk to segfault at address 0x15c.  In my own testing with
this program, I can hit that case within a few minutes.

This program requires python 2 for use with ari-py (I recommend pypy
for this purpose), as well as pjsua for use as the inbound sip caller.

It's a bit noisy - feel free to pipe pjsua stdout to open(os.devnull)
and quash requests.HTTPError

"""

import os
import ari
import time
import traceback
import subprocess
import threading
import requests.exceptions

# as configured in extensions.conf
APP_NAME = "test-application"

# as configured in http.conf
HTTP_ADDRESS = "http://172.18.0.3:8088/"

# as configured in pjsip.conf
SIP_ADDRESS = "sip:3:x@172.18.0.3:6060"

PJSUA_STDOUT=None
# uncomment for quiet.
PJSUA_STDOUT=open(os.devnull)

client = ari.connect(HTTP_ADDRESS, "admin", "peekaboo")


class SlowSnoop(object):
    done = False
    attempts = 0
    in_call = False

    def start_pjsua(self):
        self.process = subprocess.Popen([
            "pjsua", "--no-tcp", "--local-port", "5555", "--null-audio"
        ], stdin=subprocess.PIPE, stdout=PJSUA_STDOUT)

    def pjsua_line(self, text):
        self.process.stdin.write(text + "\n")
        self.process.stdin.flush()

    def place_call(self):
        self.attempts += 1
        print "attempts:", self.attempts
        self.pjsua_line("m")
        self.pjsua_line(SIP_ADDRESS)

    def hang_up(self):
        self.pjsua_line("ha")
        self.in_call = False

    def start_call_process(self):
        thread = threading.Thread(target=ss.call_process)
        thread.start()
        return thread

    def call_process(self):
        while self.in_call and not self.done:
            time.sleep(1)
        if self.done:
            return
        self.in_call = True
        time.sleep(1)
        print "-" * 60
        print "placing call..."
        self.place_call()

    def on_stasis_start(self, channel, event):
        try:
            channel = channel["channel"]
            print channel, event
            channel.answer()
            print "answered channel"
        except requests.exceptions.HTTPError as e:
            if e.response is None:
                traceback.print_exc()
            elif e.response.status_code == 404:
                print "could not find channel!"
                for line in e.response.iter_lines():
                    print line.rstrip()
                return self.start_call_process() # keep trying!
            else:
                print "unexpected response to answer", e.response.status_code
                for line in e.response.iter_lines():
                    print line.rstrip()
                return self.start_call_process() # keep trying!
        except Exception:
            traceback.print_exc()

        try:
            time.sleep(3)
            print "snooping..."
            self.start_call_process() # keep trying!
            self.hang_up()
            channel.snoopChannel(app=APP_NAME, spy="out")
            print "snoop started"
        except requests.exceptions.HTTPError as e:
            if e.response is not None and e.response.status_code == 404:
                print "channel already cleaned up"
            else:
                print "unexpected response to snoop", e.response.status_code
            for line in e.response.iter_lines():
                print line.rstrip()
        except Exception as e:
            print e.__dict__.keys()
            traceback.print_exc()


ss = SlowSnoop()
ss.start_pjsua()

client.on_channel_event("StasisStart", ss.on_stasis_start)

if __name__ == "__main__":
    time.sleep(2)
    ss.start_call_process()

    #print "now run:\n  m sip:3:x@172.18.0.3:6060"
    try:
        client.run(apps=APP_NAME)
    except KeyboardInterrupt:
        pass
    finally:
        ss.done = True
        ss.process.stdin.close()
        print "attempts:", ss.attempts
